/* 
 * Copyright 2008 Jan Schmidt <thaytan@noraisin.net>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>
#include <string.h>

#include <gst/gst.h>
#include <gst/video/video.h>
#include <gst/controller/gstcontroller.h>

#include "bgremove.h"

/* GstBgRemove signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
  ARG_ANGLE,
  ARG_NOISE_LEVEL
};

#define DEFAULT_ANGLE 20.0
#define DEFAULT_NOISE_LEVEL 2.0

static void bgremove_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void bgremove_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static gboolean bgremove_start (GstBaseTransform *trans);
static gboolean bgremove_stop (GstBaseTransform *trans);
static gboolean bgremove_get_unit_size (GstBaseTransform * btrans,
    GstCaps * caps, guint * size);
static gboolean bgremove_set_caps (GstBaseTransform * btrans,
    GstCaps * incaps, GstCaps * outcaps);
static GstFlowReturn bgremove_transform (GstBaseTransform * trans,
    GstBuffer * in, GstBuffer * out);

static GstStaticPadTemplate bgremove_src_template =
  GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_YUV ("AYUV"))
  );

static GstStaticPadTemplate bgremove_sink_template =
  GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_YUV ("AYUV"))
  );

GST_BOILERPLATE (GstBgRemove, bgremove, GstVideoFilter,
    GST_TYPE_VIDEO_FILTER);

static void
bgremove_base_init (gpointer g_class)
{
  static const GstElementDetails bgremove_details =
      GST_ELEMENT_DETAILS ("Remove background from a video stream",
      "Filter/Effect/Video",
      "Adds an alpha channel removing the background from a video stream",
      "Jan Schmidt <thaytan@noraisin.net>");
  GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);

  gst_element_class_set_details (element_class, &bgremove_details);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&bgremove_sink_template));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&bgremove_src_template));
}

static void
bgremove_class_init (GstBgRemoveClass *klass)
{
  GObjectClass *gobject_class;
  GstBaseTransformClass *trans_class;

  gobject_class = G_OBJECT_CLASS (klass);
  trans_class = (GstBaseTransformClass *) klass;

  gobject_class->set_property = bgremove_set_property;
  gobject_class->get_property = bgremove_get_property;

  g_object_class_install_property (G_OBJECT_CLASS (klass), ARG_ANGLE,
      g_param_spec_float ("angle", "Angle", "Size of the colorcube to change",
          0.0, 90.0, DEFAULT_ANGLE,
          (GParamFlags) G_PARAM_READWRITE | GST_PARAM_CONTROLLABLE));
  g_object_class_install_property (G_OBJECT_CLASS (klass), ARG_NOISE_LEVEL,
      g_param_spec_float ("noise_level", "Noise Level", "Size of noise radius",
          0.0, 64.0, DEFAULT_NOISE_LEVEL,
          (GParamFlags) G_PARAM_READWRITE | GST_PARAM_CONTROLLABLE));

  trans_class->get_unit_size = GST_DEBUG_FUNCPTR (bgremove_get_unit_size);
  trans_class->set_caps = GST_DEBUG_FUNCPTR (bgremove_set_caps);
  trans_class->start = GST_DEBUG_FUNCPTR (bgremove_start);
  trans_class->stop = GST_DEBUG_FUNCPTR (bgremove_stop);
  trans_class->transform = GST_DEBUG_FUNCPTR (bgremove_transform);
}

static void
bgremove_init (GstBgRemove *bgremove, GstBgRemoveClass *klass)
{
  GST_DEBUG ("bgremove_init");


  /* do stuff */
}

static gboolean
bgremove_get_unit_size (GstBaseTransform * btrans, GstCaps * caps,
    guint * size)
{
  GstStructure *structure;
  gboolean res = FALSE;
  gint width, height;

  structure = gst_caps_get_structure (caps, 0);
  g_return_val_if_fail (structure != NULL, FALSE);
  
  if (gst_structure_get_int (structure, "width", &width) &&
      gst_structure_get_int (structure, "height", &height)) {
    *size = gst_video_format_get_size (GST_VIDEO_FORMAT_AYUV, width, height);
    res = TRUE;
  }

  return res;
}

static gboolean
bgremove_set_caps (GstBaseTransform * btrans, GstCaps * incaps,
    GstCaps * outcaps)
{
  GstBgRemove *src = GST_BGREMOVE (btrans);
  GstStructure *structure;

  structure = gst_caps_get_structure (incaps, 0);
  g_return_val_if_fail (structure != NULL, FALSE);

  if (!gst_video_format_parse_caps (incaps, NULL, &src->width, &src->height))
    return FALSE;

  /* Clear the background on a resize. */
  gst_buffer_replace (&src->bg, NULL);

  return TRUE;
}

static void
init_angle_params (GstBgRemove *bgrem)
{
  float tmp;

  bgrem->accept_angle_cos = cos (M_PI * bgrem->angle / 180);
  bgrem->accept_angle_sin = sin (M_PI * bgrem->angle / 180);

  tmp = 15 * tan (M_PI * bgrem->angle / 180);
  tmp = MIN (tmp, 255);
  bgrem->accept_angle_tg = tmp;
  tmp = 15 / tan (M_PI * bgrem->angle / 180);
  tmp = MIN (tmp, 255);
  bgrem->accept_angle_ctg = tmp;
}

static inline void
init_params (GstBgRemove *bgrem, guint8 *target)
{
  float kgl;
  float tmp;
  float tmp_u, tmp_v;

  memcpy (bgrem->last_target, target, 4);

  bgrem->y = (float)(target[1]) - 16; 
  tmp_u = (float)(target[2]) - 128;
  tmp_v = (float)(target[3]) - 128;

  kgl = sqrt (tmp_u * tmp_u + tmp_v * tmp_v);
  bgrem->cb = 127 * (tmp_u / kgl);
  bgrem->cr = 127 * (tmp_v / kgl);

  tmp = 1 / (kgl);
  bgrem->one_over_kc = 255 * 2 * tmp - 255;
  tmp = 15 * (float) (bgrem->y) / kgl;
  tmp = MIN (tmp, 255);
  bgrem->kfgy_scale = tmp;
  bgrem->kg = MIN (kgl, 127);

#if 0
  g_print ("init: 0x%02x:0x%02x:0x%02x => %g:%d:%d:%d:%d:%d\n",
      target[1],target[2],target[3], bgrem->y, bgrem->cb,  bgrem->cr,
      bgrem->one_over_kc,  bgrem->kfgy_scale,  bgrem->kg);
#endif
}

static inline void
check_params (GstBgRemove *bgrem, guint8 *target)
{
  if (target[1] == bgrem->last_target[1] &&
      target[2] == bgrem->last_target[2] &&
      target[3] == bgrem->last_target[3])
    return; /* No change */

  /* FIXME: Caching needed */
  init_params (bgrem, target);
}

static inline void
pixel_match (GstBgRemove *bgrem, guint8 *out, guint8 *pix, guint8 *bg_data)
{
  gint b_alpha;
  gint x, z, u, v, y, a;
  gint tmp, tmp1;
  gint x1, y1;

  /* Quick check for colour proximity */
  tmp = (pix[1] - bg_data[1]);
  tmp1 = tmp * tmp;
  tmp = (pix[2] - bg_data[2]);
  tmp1 += tmp * tmp;
  tmp = (pix[3] - bg_data[3]);
  tmp1 += tmp * tmp;
  if (tmp1 <= 3*64) {
    /* Very close to the bg colour */
    *out++ = 0;
    *out++ = 0;
    *out++ = 128;
    *out++ = 128;

#if 0
    if (tmp1 < 3*9) {
      /* For really small deviations, update the bg pixel */
      bg_data[0] = pix[0];
      bg_data[1] = pix[1];
      bg_data[2] = pix[2];
      bg_data[3] = pix[3];
    }
#endif
    return;
  }
#if 0
  else {
    *out++ = *pix++;
    *out++ = *pix++;
    *out++ = *pix++;
    *out++ = *pix++;
  }
#endif

  check_params (bgrem, bg_data);

  a = pix[0];
  y = pix[1];
  u = (gint)(pix[2]) - 128;
  v = (gint)(pix[3]) - 128;

  /* Convert foreground to XZ coords where X direction is defined by
     the key color */
  tmp = ((short) u * bgrem->cb + (short) v * bgrem->cr) >> 7;
  x = CLAMP (tmp, -128, 127);
  tmp = ((short) v * bgrem->cb - (short) u * bgrem->cr) >> 7;
  z = CLAMP (tmp, -128, 127);

  /* WARNING: accept angle should never be set greater than "somewhat less
     than 90 degrees" to avoid dealing with negative/infinite tg. In reality,
     80 degrees should be enough if foreground is reasonable. If this seems
     to be a problem, go to alternative ways of checking point position
     (scalar product or line equations). This angle should not be too small
     either to avoid infinite ctg (used to suppress foreground without use of
     division) */

  tmp = ((short) (x) * bgrem->accept_angle_tg) >> 4;
  tmp = MIN (tmp, 127);

  if (abs (z) > tmp) {
    /* keep foreground Kfg = 0 */
    b_alpha = a;
  } else {
    /* Compute Kfg (implicitly) and Kbg, suppress foreground in XZ coord
       according to Kfg */
    tmp = ((short) (z) * bgrem->accept_angle_ctg) >> 4;
    tmp = CLAMP (tmp, -128, 127);
    x1 = abs (tmp);
    y1 = z;

    tmp1 = x - x1;
    tmp1 = MAX (tmp1, 0);
    b_alpha = (((unsigned char) (tmp1) *
            (unsigned short) (bgrem->one_over_kc)) / 2);
    b_alpha = 255 - CLAMP (b_alpha, 0, 255);
    b_alpha = (a * b_alpha) >> 8;

    tmp = ((unsigned short) (tmp1) * bgrem->kfgy_scale) >> 4;
    tmp1 = MIN (tmp, 255);

    tmp = y - tmp1;
    y = MAX (tmp, 0);

    /* Convert suppressed foreground back to CbCr */
    tmp = ((char) (x1) * (short) (bgrem->cb) -
        (char) (y1) * (short) (bgrem->cr)) >> 7;
    u = CLAMP (tmp, -128, 127);

    tmp = ((char) (x1) * (short) (bgrem->cr) +
        (char) (y1) * (short) (bgrem->cb)) >> 7;
    v = CLAMP (tmp, -128, 127);

    /* Deal with noise. For now, a circle around the key color with
       radius of noise_level treated as exact key color. Introduces
       sharp transitions.
     */
    tmp = z * (short) (z) + (x - bgrem->kg) * (short) (x - bgrem->kg);
    tmp = MIN (tmp, 0xffff);

    if (tmp < bgrem->noise_level * bgrem->noise_level) {
      b_alpha = 0;
    }
  }

  *out++ = b_alpha;
#if 0 /* Paint alpha only */
  *out++ = b_alpha;
  *out++ = 128;
  *out++ = 128;
#else /* Paint pixel */
  *out++ = pix[1];
  *out++ = pix[2];
  *out++ = pix[3];
#endif

#if 0
  g_print ("0x%02x:0x%02x:0x%02x => 0x%02x:0x%02x:0x%02x -> %u\n",
    bg_data[1], bg_data[2], bg_data[3], pix[1], pix[2], pix[3], b_alpha);
#endif

}

static void
process_ayuv (GstBgRemove *bgrem, GstBuffer *in, GstBuffer *out)
{
  int y, x;
  guint8 *in_line, *in_data;
  guint8 *out_line, *out_data;
  guint8 *bg_line, *bg_data;
  gint stride =
      gst_video_format_get_row_stride (GST_VIDEO_FORMAT_AYUV, 0, bgrem->width);

  out_line = GST_BUFFER_DATA (out);
  in_line = GST_BUFFER_DATA (in);
  bg_line = GST_BUFFER_DATA (bgrem->bg);

  for (y = 0; y < bgrem->height; y++) {
    out_data = out_line;
    in_data = in_line;
    bg_data = bg_line;

    for (x = 0; x < bgrem->width; x++) {
      /* Correlate bg buffer with input and copy to output */
      pixel_match (bgrem, out_data, in_data, bg_data);

      out_data += 4;
      in_data += 4;
      bg_data += 4;
    }

    out_line += stride;
    in_line += stride;
    bg_line += stride;
  }
}

#if 0
static void
acquire_bg (GstBgRemove *bgrem, GstBuffer *in)
{
  int y, x;
  guint8 *in_line, *in_data;
  guint8 *bg_line, *bg_data;
  gint stride;
  
  if (bgrem->bg == NULL) {
    /* No existing background - take the first frame, as our reference */
    bgrem->bg = gst_buffer_copy (in);
    return;
  }

  /* Average the incoming frame into the bg */
  stride =
      gst_video_format_get_row_stride (GST_VIDEO_FORMAT_AYUV, 0, bgrem->width);
  in_line = GST_BUFFER_DATA (in);
  bg_line = GST_BUFFER_DATA (bgrem->bg);

  for (y = 0; y < bgrem->height; y++) {
    in_data = in_line;
    bg_data = bg_line;

    for (x = 0; x < bgrem->width; x++) {
      guint16 val;

      val = ((guint16) in_data[1] + ((guint16)(bg_data[1]) * 7)) / 8; /* Y */
      bg_data[1] = (guint8) val;
      val = ((guint16) in_data[2] + ((guint16)(bg_data[2]) * 7)) / 8; /* U */
      bg_data[2] = (guint8) val;
      val = ((guint16) in_data[3] + ((guint16)(bg_data[3]) * 7)) / 8; /* V */
      bg_data[3] = (guint8) val;

      in_data += 4;
      bg_data += 4;
    }

    in_line += stride;
    bg_line += stride;
  }
}
#else
static void
acquire_bg (GstBgRemove *bgrem, GstBuffer *in)
{
  if (bgrem->buffer_num < 10) {
    if (bgrem->bg)
      gst_buffer_unref (bgrem->bg);
    bgrem->bg = gst_buffer_copy (in);
  }
}
#endif

static gboolean
bgremove_start (GstBaseTransform *trans)
{
  GstBgRemove *bgrem = GST_BGREMOVE (trans);

  init_angle_params (bgrem);
  init_params (bgrem, bgrem->last_target);

  return TRUE;
}

static gboolean
bgremove_stop (GstBaseTransform *trans)
{
  GstBgRemove *bgrem = GST_BGREMOVE (trans);

  gst_buffer_replace (&bgrem->bg, NULL);

  return TRUE;
}

static GstFlowReturn
bgremove_transform (GstBaseTransform * trans, GstBuffer * in, GstBuffer * out)
{
  GstFlowReturn ret = GST_FLOW_OK;
  GstBgRemove *bgrem = GST_BGREMOVE (trans);

  acquire_bg (bgrem, in);
  process_ayuv (bgrem, in, out);
  
  bgrem->buffer_num++;

  return ret;
}

static void
bgremove_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstBgRemove *bgrem;

  g_return_if_fail (GST_IS_BGREMOVE (object));
  bgrem = GST_BGREMOVE (object);

  GST_DEBUG ("bgremove_set_property");
  switch (prop_id) {
    case ARG_ANGLE:
      bgrem->angle = g_value_get_float (value);
      init_angle_params (bgrem);
      init_params (bgrem, bgrem->last_target);
      break;
    case ARG_NOISE_LEVEL:
      bgrem->noise_level = g_value_get_float (value);
      init_angle_params (bgrem);
      init_params (bgrem, bgrem->last_target);
      break;
#if 0
    case ARG_METHOD:
      bgrem->method = g_value_get_enum (value);
      break;
#endif
    default:
      break;
  }
}

static void
bgremove_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  GstBgRemove *bgrem;

  g_return_if_fail (GST_IS_BGREMOVE (object));
  bgrem = GST_BGREMOVE (object);

  switch (prop_id) {
    case ARG_ANGLE:
      g_value_set_float (value, bgrem->angle);
      break;
    case ARG_NOISE_LEVEL:
      g_value_set_float (value, bgrem->noise_level);
      break;
#if 0
    case ARG_METHOD:
      g_value_set_enum (value, bgrem->method);
      break;
#endif
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

GST_DEBUG_CATEGORY (bgremove_debug);
#define GST_CAT_DEFAULT bgremove_debug

static gboolean
plugin_init (GstPlugin * plugin)
{
  gboolean result = TRUE;

  GST_DEBUG_CATEGORY_INIT (bgremove_debug, "bgremove elements",
      0, "Background remove (video to alpha stream)");

  result &= gst_element_register (plugin, "bgremove",
                GST_RANK_PRIMARY, GST_TYPE_BGREMOVE);

  return result;
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "bgremove",
    "Background remove (video to alpha stream)",
    plugin_init, VERSION, "LGPL", "GStreamer", "http://gstreamer.net/")
